/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  Routines for processing user feedback (provided via the pitch bend wheel)
 */


#include "defs.h"


/* how far back to apply user feedback, in half seconds (3 = 1.5 secs) */
#define BACKLOG     3


/* a list of all the patterns that have recently been played */
PATTERN *feedback_patterns[BACKLOG][MAX_PATTERNS];



/* adjust_pattern:
 *  Helper function to adjust the weighting of a pattern,
 */
void adjust_pattern(p, adjustment, adjust_mutator)
PATTERN *p;
long adjustment;
int adjust_mutator;
{
    PATTERN *pt;
    MUTATOR *m;
    int c;

    if (!p)
	return;

    /* find and adjust the item and its immediate descendants */
    for (c=0; the_patterns[c].data; c++) {
	pt = (PATTERN *)the_patterns[c].data;

	if ((pt == p) || (pt->parent == p)) {

	    the_patterns[c].coolness += adjustment;

	    if (the_patterns[c].coolness < 0)
		the_patterns[c].coolness = 0;
	    else if (the_patterns[c].coolness > MAX_PROB)
		the_patterns[c].coolness = MAX_PROB;
	}
    }

    if (adjust_mutator) {
	m = p->mutator;
	if (m) {
	    /* adjust the mutator that created the pattern */
	    adjust_weighting(the_mutators, (long)m, adjustment/8);
	    if (m->integer_list) {
		/* and adjust the integer mutator parameter */
		adjust_weighting(m->integer_list, p->mutator_i, adjustment/8);
	    }
	}
    }
}



/* do_feedback:
 *  Called twice a second to process user feedback, provided via the
 *  pitch bend wheel.
 */
void do_feedback()
{
    int c1, c2;
    PATTERN *p;

    /* shift backlog tables along, and copy playing patterns into them */
    for (c1=BACKLOG-1; c1>0; c1--) 
	for (c2=0; c2<MAX_PATTERNS; c2++)
	    feedback_patterns[c1][c2] = feedback_patterns[c1-1][c2];

    for (c1=0; c1<MAX_PATTERNS; c1++)
	feedback_patterns[0][c1] = playing_pattern[c1].pattern;

    /* adjust the active patterns and their immediate parents */
    if (pitch_bend != 0) {
	for (c1=0; c1<BACKLOG; c1++) {
	    for (c2=0; c2<MAX_PATTERNS; c2++) {
		p = feedback_patterns[c1][c2];
		if (p) {
		    adjust_pattern(p, pitch_bend, (c1==1));
		    if (p->parent)
			adjust_pattern(p->parent, pitch_bend/2, (c1==1));
		}
	    }

	    /* keep midi playback going! */
	    do_midi();
	}
    }
}



/* feedback_del_pattern:
 *  Called whenever a pattern is deleted, to remove any pointers to it from
 *  the backlog list.
 */
void feedback_del_pattern(p)
PATTERN *p;
{
    int c1, c2;

    for (c1=0; c1<BACKLOG; c1++)
	for (c2=0; c2<MAX_PATTERNS; c2++)
	    if (feedback_patterns[c1][c2] == p)
		feedback_patterns[c1][c2] = NULL;
}



/* feedback_init:
 *  Sets up the feedback manager ready for use.
 */
void feedback_init()
{
    int c1, c2;

    for (c1=0; c1<BACKLOG; c1++)
	for (c2=0; c2<MAX_PATTERNS; c2++)
	    feedback_patterns[c1][c2] = NULL;
}

