/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  Initialisation and main control loop.
 */


#include "defs.h"


/* somewhere to put unwanted return values */
long nowhere;


/* should we print debugging information? */
int trace = FALSE;


/* freeze time to prevent errors accumulating while we process things */
long my_clock = 0;


/* This is a list of processes that need perfoming occassionally.
 * The main control loop steps through this list once per second, calling
 * the functions in order (so a function which occurs once in the list
 * will be called once per second, one which occurs twice will be called
 * twice, etc.
 */

#define PROCESS_COUNT   5

typedef void (FUNCTION)();

FUNCTION *process[PROCESS_COUNT] =
{
    do_feedback,
    mutate,
    do_feedback,
    kill_patterns,
    worsen_with_age
};



/* usage:
 *  Displays a list of command line options.
 */
void usage()
{
    printf("\nUsage: pp [<filename>] [trace] [-c<x>] [-n<x>] [-h]\n\n");
    printf("    <filename>        read/write mutator probabilities from this file\n");
    printf("    -o<filename.mid>  log the output to this MIDI file\n");
    printf("    trace             send trace information to stderr\n");
    printf("    -c<x>             play metronome on MIDI channel <x> (default 10)\n");
    printf("    -n<x>             play metronome on note number <x> (default 42)\n");
    printf("    -h                display this message\n\n");
}



int main(argc, argv)
int argc;
char *argv[];
{
    int finish = FALSE;
    int process_pos = 0;
    long last_time = 0;
    char fname[80] = "";
    int c;

    /* process command line options */
    for (c=1; c<argc; c++) {
	if (stricmp(argv[c], "trace") == 0) {
	    trace = TRUE;
	}
	else if ((argv[c][0] == '/') || (argv[c][0] == '-')) {
	    if ((argv[c][1] == 'c') || (argv[c][1] == 'C')) {
		metronome_channel = atoi(argv[c]+2) - 1;
	    }
	    else if ((argv[c][1] == 'n') || (argv[c][1] == 'N')) {
		metronome_note = atoi(argv[c]+2);
	    }
	    else if ((argv[c][1] == 'o') || (argv[c][1] == 'O')) {
	       log_name = argv[c]+2;
	    }
	    else {
		if ((argv[c][1] != 'h') && (argv[c][1] != 'H'))
		    printf("\nUnknown option '%s'\n", argv[c]);
		usage();
		return 1; 
	    }
	}
	else
	    strcpy(fname, argv[c]);
    }

    if (fname[0])
	load_weightings(fname);

    if (hw_init() != 0) {
	printf("\nError initialising MIDI hardware!\n");
	return 1;
    }

    if (log_name) {
	log_file = fopen(log_name, "wb");

	if (!log_file) {
	    printf("\nError writing '%s'\n", log_name);
	    hw_cleanup();
	    return 1;
	}

	midi_puts("MThd");
	midi_putl(6); 
	midi_putw(0); 
	midi_putw(1); 
	midi_putw(120); 
	midi_puts("MTrk");
	midi_putl(0); 

	log_len = 0;
    }

    ui_init();
    midi_init();
    pattern_init();
    feedback_init();

    srandom(time(NULL));

    do {
	/* process MIDI thru, playback and recording */
	do_midi();

	/* handle any input from the user */
	do_ui(&finish);

	/* handle periodic, scheduled processes */
	if (my_clock > (last_time + CLK_TCK/PROCESS_COUNT)) {
	    (*process[process_pos++])();
	    if (process_pos >= PROCESS_COUNT)
		process_pos = 0;
	    last_time = my_clock;
	}

    } while (!finish);

    pattern_cleanup();
    midi_cleanup();
    ui_cleanup();
    hw_cleanup();

    if (fname[0]) {
	save_weightings(fname);
	screen_out("Mutator probabilities written into '");
	screen_out(fname);
	screen_out("'\r\n\r\n");
    }

    if (log_file) {
	fseek(log_file, 18, 0);
	midi_putl(log_len);

	fclose(log_file);
	screen_out("MIDI log file written into '");
	screen_out(log_name);
	screen_out("'\r\n\r\n");
    }

    return 0;
}



#ifndef DJGPP


/* stricmp:
 *  For some reason there is no stricmp() in the Mark Williams C library,
 *  so I need to provide one of my own.
 */
int stricmp(s1, s2)
char *s1;
char *s2;
{
    char c1, c2;

    do { 
	c1 = *(s1++);
	if ((c1 >= 'a') && (c1 <= 'z'))
	    c1 = c1 - 'a' + 'A';

	c2 = *(s2++);
	if ((c2 >= 'a') && (c2 <= 'z'))
	    c2 = c2 - 'a' + 'A';

	if (c1 != c2)
	    return 1;

    } while (c1);

    return 0;
}


#endif

