/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  Routines for manipulating pattern structures
 */


#include "defs.h"


/* global pool of patterns (the data field points to the pattern structure).
 * This is twice the size of the MAX_PATTERNS, so we can handle temporarily
 * storing more than the maximum before we get round to deleting old ones.
 */
LIST_MEMBER the_patterns[MAX_PATTERNS*2];


/* the patterns that are currently playing */
PLAYING_PATTERN playing_pattern[MAX_PATTERNS];


/* flags that this pattern should be triggered on the next down-beat */
PATTERN *waiting_to_play = NULL;


/* the larger this value, the more patterns will be played at a time */
int pattern_density = 8;



/* resize_pattern:
 *  Wrapper for realloc, which only expands a pattern if neccessary.
 */
PATTERN *resize_pattern(p, length)
PATTERN *p;
int length;
{
    int s;

    if (p->allocated_size >= length)
	return p;

    s = length + 64;

    p = realloc(p, (int)(sizeof(PATTERN) + sizeof(NOTE) * s));
    if (!p)
	return NULL;

    p->allocated_size = s;
    return p;
}



/* play_pattern:
 *  Starts playback of a pattern.
 */
void play_pattern(p)
PATTERN *p;
{
    int x;

    /* find a free slot in the playing pattern array */
    for (x=0; x<MAX_PATTERNS; x++) {
	if (playing_pattern[x].pattern == NULL) {
	    playing_pattern[x].pos = 0;
	    playing_pattern[x].start_time = my_clock;
	    playing_pattern[x].pattern = p;
	    return;
	}
    }
}



/* destroy_pattern:
 *  Destroys a pattern structure when it is no longer required.
 */
void destroy_pattern(p)
PATTERN *p;
{
    int x;

    TRACE("destroy_pattern()\n");

    if (!p)
	return;

    /* make sure that the pattern isn't currently playing */
    for (x=0; x<MAX_PATTERNS; x++) {
	if (playing_pattern[x].pattern == p)
	    playing_pattern[x].pattern = NULL;

	if ((playing_pattern[x].pattern) && 
	    (playing_pattern[x].pattern->parent == p))
	    playing_pattern[x].pattern->parent = NULL;
    }

    if (waiting_to_play == p)
	waiting_to_play = NULL;

    feedback_del_pattern(p);

    /* and free the memory */
    free(p);
}



/* pattern_init:
 *  Initialises the pattern manager at program startup.
 */
void pattern_init()
{
    int x;

    for (x=0; x<MAX_PATTERNS; x++) {
	playing_pattern[x].pattern = NULL;
	playing_pattern[x].pos = 0;
	playing_pattern[x].start_time = 0;
    } 

    for (x=0; x<MAX_PATTERNS*2; x++) {
	the_patterns[x].data = 0;
	the_patterns[x].coolness = 0;
    }
}



/* destroy_all_patterns:
 *  Empties the list of recorded patterns.
 */
void destroy_all_patterns()
{
    int x;

    for (x=0; x<MAX_PATTERNS*2; x++) {
	if (the_patterns[x].data) {
	    destroy_pattern((PATTERN *)the_patterns[x].data);
	    the_patterns[x].data = 0;
	    the_patterns[x].coolness = 0;
	}
    }
}



/* pattern_cleanup:
 *  Shuts down the pattern manager at program termination.
 */
void pattern_cleanup()
{
    destroy_all_patterns();
}



/* add_pattern:
 *  Adds a newly recorded pattern to the list of available material.
 */
void add_pattern(p, coolness)
PATTERN *p;
long coolness;
{
    if (!p)
	return;

    if ((p->length > 0) && (get_list_size(the_patterns) < MAX_PATTERNS*2)) {
	add_to_list(the_patterns, (long)p, coolness);
	waiting_to_play = p;
    }
    else
	destroy_pattern(p);
}



/* play_patterns:
 *  Called periodically to check how many patterns are playing, and start 
 *  up new ones as required.
 */
void play_patterns()
{
    int x, c;
    PATTERN *p;

    c = 0;
    for (x=0; x<MAX_PATTERNS; x++)
	if (playing_pattern[x].pattern)
	    c++;

    x = pattern_density - 7;

    while (PROB(25)) 
	x++;

    while (PROB(25)) 
	x--;

    if ((c < x) && (waiting_to_play)) {
	play_pattern(waiting_to_play);
	waiting_to_play = NULL;
	c++;
    }

    while (c++ < x) {
	p = (PATTERN *)pick_from_list(the_patterns, &nowhere);
	if (p)
	    play_pattern(p);
    }
}



/* kill_patterns:
 *  Called periodically to check how many patterns exist, and delete
 *  old/unpopular ones if there are too many.
 */
void kill_patterns()
{
    int c = get_list_size(the_patterns);
    PATTERN *p;

    while (c > MAX_PATTERNS) {
	p = (PATTERN *)pick_bad_item(the_patterns);
	if (p) {
	    remove_from_list(the_patterns, (long)p);
	    destroy_pattern(p);
	}
	c--;
    }
}



/* worsen_with_age:
 *  Called periodically to make old patterns less popular.
 */
void worsen_with_age()
{
    int x;

    for (x=0; x<MAX_PATTERNS*2; x++)
	if (the_patterns[x].coolness > 1)
	    the_patterns[x].coolness--;
}
