/*
 *  MIDI pattern processor
 *  by Shawn Hargreaves, 1995/96
 *
 *  Mutation routines for randomly displacing notes within a pattern, and
 *  for compressing patterns to fit within a specified range of pitches.
 */


#include "defs.h"


MUTATE_FUNC displace_func, range_changer_func, thinner_func, thickener_func;


/* various different types of note displacement */

#define DISPLACE_OCTAVES            1
#define DISPLACE_RANDOM             2
#define DISPLACE_MODAL              3
#define DISPLACE_SWAP               4


LIST_MEMBER displace_params[] =
{
    /* <description>                <data>                        <coolness> */
    { "Octave Displacement",        DISPLACE_OCTAVES,             MAX_PROB/4 },
    { "Random Displacement",        DISPLACE_RANDOM,              MAX_PROB/4 },
    { "Modal Displacement",         DISPLACE_MODAL,               MAX_PROB/4 },
    { "Swap Pitches",               DISPLACE_SWAP,                MAX_PROB/4 },
    { NULL, 0, 0 }
};



MUTATOR displace_notes =
{
    displace_func,              /* displaces notes within a pattern */
    displace_params             /* what sort of displacement? */
};



/* various different types of range alteration */

#define RANGE_TO_OCTAVE             1
#define RANGE_STRETCH               2
#define RANGE_MODAL_STRETCH         3


LIST_MEMBER range_change_params[] =
{
    /* <description>                <data>                        <coolness> */
    { "Range to Octave",            RANGE_TO_OCTAVE,              MAX_PROB/4 },
    { "Stretch Range",              RANGE_STRETCH,                MAX_PROB/4 },
    { "Modal Stretch",              RANGE_MODAL_STRETCH,          MAX_PROB/4 },
    { NULL, 0, 0 }
};



MUTATOR range_changer =
{
    range_changer_func,         /* alter the pitch range of a pattern */
    range_change_params         /* what sort of change? */
};



MUTATOR thinner =
{
    thinner_func,               /* removes notes from a pattern */
    NULL                        /* don't need an integer parameter */
};



/* various different types of thickening */

#define THICKEN                     1
#define THICKEN_MODAL               2


LIST_MEMBER thickener_params[] =
{
    /* <description>                <data>                        <coolness> */
    { "Random Thicken",             THICKEN,                      MAX_PROB/4 },
    { "Modal Thicken",              THICKEN_MODAL,                MAX_PROB/4 },
    { NULL, 0, 0 }
};



MUTATOR thickener =
{
    thickener_func,             /* adds notes to a pattern */
    thickener_params            /* what sort of thickening? */
};



/* dispace_func:
 *  Displaces notes within a pattern.
 */
PATTERN *displace_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    int c, x, t;

    p = clone_pattern(p1);

    if (p) {
	if (i == DISPLACE_OCTAVES) {
	    TRACE("displace_func(DISPLACE_OCTAVES)\n");

	    for (c=0; c<p->length; c++)
		if (PROB(25))
		    p->note[c].pitch += (PROB(50) ? 12 : -12);
	}
	else if (i == DISPLACE_SWAP) {
	    TRACE("displace_func(DISPLACE_SWAP)\n");

	    for (c=0; c<p->length; c++) {
		if (PROB(25)) {
		    x = random() % p->length;
		    t = p->note[c].pitch;
		    p->note[c].pitch = p->note[x].pitch;
		    p->note[x].pitch = t;
		}
	    }
	} 
	else {
	    if (i == DISPLACE_MODAL) {
		TRACE("displace_func(DISPLACE_MODAL)\n");
		fill_temp_with_mode(p);
	    }
	    else
		TRACE("displace_func(DISPLACE_RANDOM)\n");

	    for (c=0; c<p->length; c++) {
		if (PROB(25)) {
		    p->note[c].pitch += (random() % 9) - 4;
		    if (i == DISPLACE_MODAL)
			p->note[c].pitch = closest_note_in_mode(p->note[c].pitch);
		}
	    }
	}
    }

    return p;
}



/* range_changer_func:
 *  Changes the pitch range of a pattern.
 */
PATTERN *range_changer_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    int c;
    int middle;
    float scale;

    p = clone_pattern(p1);

    if (p) {
	middle = median_pitch(p);

	if (i == RANGE_TO_OCTAVE) {
	    TRACE("range_changer_func(RANGE_TO_OCTAVE)\n"); 

	    for (c=0; c<p->length; c++) {
		while (p->note[c].pitch < middle-6)
		    p->note[c].pitch += 12;

		while (p->note[c].pitch > middle+6)
		    p->note[c].pitch -= 12;
	    }
	}
	else {
	    if (i == RANGE_MODAL_STRETCH) {
		TRACE("range_changer_func(RANGE_MODAL_STRETCH)\n"); 
		fill_temp_with_mode(p);
	    }
	    else
		TRACE("range_changer_func(RANGE_STRETCH)\n"); 

	    scale = 0.5 + (float)(random() % 75) / 50.0;

	    for (c=0; c<p->length; c++) {
		p->note[c].pitch = middle + 
			    (int)((float)(p->note[c].pitch - middle) * scale);

		if (i == RANGE_MODAL_STRETCH)
			p->note[c].pitch = closest_note_in_mode(p->note[c].pitch);
	    }
	}
    }

    return p;
}



/* thinner_func:
 *  Removes notes from a pattern.
 */
PATTERN *thinner_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    int c, x;

    TRACE("thinner_func()\n"); 

    p = clone_pattern(p1);

    if ((p) && (p->length >= 3)) {
	c = random() % (p->length/3);

	/* remove randomly chosen notes */
	while ((c > 0) && (p->length > 1)) {
	    for (x=random() % p->length; x<p->length-1; x++)
		p->note[x] = p->note[x+1];

	    p->length--;
	    c--;
	}
    }

    return p;
}



/* thickener_func:
 *  Adds new notes to a pattern.
 */
PATTERN *thickener_func(p1, p2, i)
PATTERN *p1;
PATTERN *p2;
int i;
{
    PATTERN *p;
    int len;
    int pos, c;

    if (p1->length < 3)
	return NULL;

    len = p1->length + (random() % p1->length/2);
    p = malloc((int)(sizeof(PATTERN) + len * sizeof(NOTE)));

    if (p) {
	if (i == THICKEN_MODAL) {
	    TRACE("thickener_func(THICKEN_MODAL)\n");
	    fill_temp_with_mode(p1);
	}
	else
	    TRACE("thickener_func(THICKEN)\n");

	p->channel = p1->channel;
	p->length = p1->length;
	p->allocated_size = len; 

	for (c=0; c<p->length; c++)
	    p->note[c] = p1->note[c];

	/* add notes at randomly chosen positions */
	while (p->length < len) {
	    pos = 1 + random() % (p->length-2);

	    /* shift existing notes along to make room */
	    for (c=p->length; c>pos; c--)
		p->note[c] = p->note[c-1];

	    p->length++;

	    /* insert new pitch */
	    if (PROB(50))
		p->note[pos].pitch = p->note[pos-1].pitch + 1 + (random() % 7);
	    else
		p->note[pos].pitch = p->note[pos-1].pitch - 1 - (random() % 7);

	    if (i == THICKEN_MODAL)
		p->note[pos].pitch = closest_note_in_mode(p->note[pos].pitch);

	    p->note[pos].vel = p->note[pos-1].vel;

	    /* and new time */
	    c = p->note[pos+1].time - p->note[pos-1].time;
	    if (c > 0)
		p->note[pos].time = p->note[pos-1].time + (random() % c);
	    else
		p->note[pos].time = p->note[pos-1].time;

	    /* length not known yet: we can sort this out later */
	    p->note[pos].length = 0;
	}

	quantize(p);

	/* sort out the lengths of new notes */
	for (c=1; c<p->length-1; c++) {
	    if (p->note[c].length == 0) {
		if (p->note[c].time == p->note[c-1].time)
		    p->note[c].length = p->note[c-1].length;
		else if (p->note[c].time == p->note[c+1].time)
		    p->note[c].length = p->note[c+1].length;
		else {
		    p->note[c].length = p->note[c+1].time - p->note[c].time;
		    p->note[c-1].length = p->note[c].time - p->note[c-1].time;
		}
	    }
	}
    }

    return p;
}

